#define _GNU_SOURCE 

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <locale.h>
#include <stdint.h>

#include <libxml/encoding.h>
#include <libxml/xmlwriter.h>
#include <curl/curl.h>

#include "strata_client.h"

struct reportfile {
  char* tmpdir;
  xmlTextWriterPtr writer;
  xmlBufferPtr buf;
  int issignature;
};

static char* internal_error_string = NULL;

static char*
ssprintf( const char* format, ... ) {
  size_t r;
  va_list p;
  char* retval;

  va_start(p, format);
  r = vasprintf(&retval, format, p);
  va_end(p);

  if (r < 0) {
    internal_error_string = strdup("ssprintf: vasprintf failed");
    if (!internal_error_string) 
      internal_error_string = "ssprintf: vasprintf failed";
    return 0;
  }

  return retval;
}

static void
internal_error_printf( const char* format, ... ) {
  size_t r;
  va_list p;

  va_start(p, format);
  r = vasprintf(&internal_error_string, format, p);
  va_end(p);

  if (r < 0) {
    internal_error_string = strdup("internal_error_printf: vasprintf failed");
    if (!internal_error_string) 
      internal_error_string = "internal_error_printf: vasprintf failed";
  } 
}

static int 
run_command( const char *cmd )
{
    int retcode = system(cmd);
    if (retcode) {
      internal_error_printf("'%s' exited with status %d", cmd, retcode);
      return -1;
    }
    return 0;
}

static int
copy_file( const char *src_name, const char *dst_name )
{
  char* cmd = ssprintf("cp %s %s", src_name, dst_name);
  if (!cmd)
    return -1;
  return run_command(cmd);
}

static char*
concat_path_file(const char *path, const char *filename) {
  // This adds concats the two names, adding or removing one slash
  // as necessary.  It does not handle the case that path ends
  // in, or filename starts with, multiple slashs.

  const char* path_end = path + strlen(path) - 1;
  if ((*path_end == '/') != (*filename == '/'))
    // one has it, the other doesn't
    return ssprintf("%s%s", path, filename);
  else if (*filename == '/')
    // they both have it
    return ssprintf("%s%s", path, filename+1);
  else
    // neither has it
    return ssprintf("%s/%s", path, filename);
}



static int
write_rgn_to_file( const char *value, size_t size,
                   const char *path, const char *fname ) {

  char* ofile_name = concat_path_file(path, fname);
  if (!ofile_name)
    return -1;

  FILE *ofile = fopen(ofile_name, "w");
  if (!ofile) {
    internal_error_printf("Can't open '%s'", ofile_name);
    return -1;
  }

  size_t value_size = (size == SIZE_MAX) ? strlen(value) : size;
  if (value_size != fwrite(value, 1, value_size, ofile)) {
    internal_error_printf("Can't write to '%s'", ofile_name);
    return -1;
  }

  fclose(ofile);
  return 0;
}

const char* 
strata_client_strerror( void ) {
  return internal_error_string;
}

//
// The purpose of all the reportfile_XXX functions is to 
// collect up the contents of a report or signature into a 
// single tarball or xml file.
// 


// 
// This allocates a reportfile_t structure and initializes it.
//
//   Set 'issignature' to true if you only want an xml file.
//   If 'issignature' is true, then any bindings which would 
//   be written to an external file are dropped.
//
reportfile_t* 
reportfile_start( int issignature ) {

  // create a new reportfile_t
  reportfile_t* file = (reportfile_t*)malloc(sizeof(reportfile_t));

  file->issignature = issignature;

  // make a temp directory to work in
  file->tmpdir = strdup("/tmp/reportfileXXXXXX");
  if (mkdtemp(file->tmpdir) == NULL) {
    internal_error_printf("Can't mkdir a temporary directory in /tmp");
    return NULL;
  }

  // create a 'content' dir within temp directory
  char* content_dir_name = concat_path_file(file->tmpdir, "content");
  if (!content_dir_name)
    return NULL;
  if (mkdir(content_dir_name, 0700)) {
    internal_error_printf("Can't mkdir '%s'\n", content_dir_name);
    return NULL;
  }

  // set up a libxml 'buffer' and 'writer' to that buffer
  file->buf = xmlBufferCreate();
  if (file->buf == NULL) {
    internal_error_printf("strata_client: Error creating the xml buffer\n");
    return NULL;
  }
  file->writer = xmlNewTextWriterMemory(file->buf, 0);
  if (file->writer == NULL) {
    internal_error_printf("strata_client: Error creating the xml writer\n");
    return NULL;
  }

  // start a new xml document
  int rc;
  rc = xmlTextWriterStartDocument(file->writer, NULL, NULL, NULL);
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterStartDocument\n");
    return NULL;
  }

  // start the document element 'report'
  rc = xmlTextWriterStartElement(file->writer, (const xmlChar*)"report");
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterStartElement\n");
    return NULL;
  }

  // add the 'xmlns' attribue
  rc = xmlTextWriterWriteAttribute(file->writer, (const xmlChar*)"xmlns",
                                   (const xmlChar*)"http://www.redhat.com/gss/strata");
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterWriteAttribute\n");
    return NULL;
  }

  free(content_dir_name);
  return file;
}


static int
internal_reportfile_start_binding( reportfile_t* file, const char* name, 
                                   int isbinary, const char* filename ) {
  int rc;
  
  // open a binding element
  rc = xmlTextWriterStartElement(file->writer, (const xmlChar*)"binding");
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterStartElement\n");
    return -1;
  }

  // add the 'name' attribue
  rc = xmlTextWriterWriteAttribute(file->writer, (const xmlChar*)"name", (const xmlChar*)name);
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterWriteAttribute\n");
    return -1;
  }


  // add the 'filename' attribute if neede
  if (filename) {
    rc = xmlTextWriterWriteAttribute(file->writer, (const xmlChar*)"fileName", (const xmlChar*)filename);
    if (rc < 0) {
      internal_error_printf
	("strata_client: Error at xmlTextWriterWriteAttribute\n");
      return -1;
    }
  }

  // add the 'type' attribute
  if (isbinary) {
    rc = xmlTextWriterWriteAttribute(file->writer, (const xmlChar*)"type", (const xmlChar*)"binary");
    if (rc < 0) {
      internal_error_printf
	("strata_client: Error at xmlTextWriterWriteAttribute\n");
      return -1;
    }
  }
  else {
    rc = xmlTextWriterWriteAttribute(file->writer, (const xmlChar*)"type", (const xmlChar*)"text");
    if (rc < 0) {
      internal_error_printf
	("strata_client: Error at xmlTextWriterWriteAttribute\n");
      return -1;
    }
  }
  return 0;
}

//
// Add a new binding to a report, for a binding whos value is represented
//   as either a null-terminated string or start+len region of memory.
//
//   'file' is the reportfile to add the binding to.
//   'name' is the name of the binding
//   'isbinary' should be true if the data in local_filename is not UTF-8
//       (or compatible, like ASCII).
//   'filename' is the original name of the file this data was obtained from
//   
//   'value' and 'size' hold the value to be bound to 'name'
//      If 'size' == SIZE_MAX then 'value' is treated as a null-terminated
//        string; otherwise 'value' the start of a region of memory 'size' 
//        bytes long.
//
//      Note that while it is valid to set 'isbinary' for a value that is
//        a null-terminated string, it probably is not correct.
//
static int
reportfile_add_binding_from_region( reportfile_t* file, 
                                    const char* name, 
                                    const void* value,
                                    size_t size,
                                    int isbinary, 
                                    const char* filename ) {
  int rc;

  // start the binding element with primary attributes
  rc = internal_reportfile_start_binding(file, name, isbinary, filename);
  if (rc < 0)
    return rc;


  // write the value of the binding, in the way that makes sense
  //   given it's value
  if (isbinary) {
    char* content_dir_name = concat_path_file(file->tmpdir, "content");
    if (!content_dir_name)
      return -1;
    rc = write_rgn_to_file(value, size, content_dir_name, name);
    if (rc < 0)
      return rc;
    free(content_dir_name);

    char* href_name = concat_path_file("content", name);
    if (!href_name)
      return -1;
    rc = xmlTextWriterWriteAttribute(file->writer, 
                                     (const xmlChar*)"href", 
                                     (const xmlChar*)content_dir_name);
    free(href_name);
    if (rc < 0) {
      internal_error_printf
	("strata_client: Error at xmlTextWriterWriteAttribute\n");
      return -1;
    }
  }
  else {
    rc = xmlTextWriterWriteAttribute(file->writer, 
                                     (const xmlChar*)"value", 
                                     (const xmlChar*)value);
      if (rc < 0) {
        internal_error_printf
	  ("strata_client: Error at xmlTextWriterWriteAttribute\n");
        return -1;
    }
  }

  rc = xmlTextWriterEndElement(file->writer);
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterEndElement\n");
    return -1;
  }
  return 0;
}

// Add a new binding to a report, for a binding whos value is represented
//   as a null-terminated string.
//
//   'file' is the reportfile to add the binding to.
//   'name' is the name of the binding
//   'isbinary' should be true if the data in local_filename is not UTF-8
//       (or compatible, like ASCII).
//   'filename' is the original name of the file this data was obtained from
//   'value' is a null-terminated string to bound to name.
//      Note that while it is valid to set 'isbinary' for a value that is
//        a null-terminated string, it probably is not correct.
//
int
reportfile_add_binding_from_string( reportfile_t* file, 
                                    const char* name, 
                                    const char* value,
                                    int isbinary, 
                                    const char* filename ) {
  return reportfile_add_binding_from_region(file, name, 
                                            value, SIZE_MAX, 
                                            isbinary, filename);
}

// Add a new binding to a report, for a binding whos value is represented
//   as a null-terminated string.
//
//   'file' is the reportfile to add the binding to.
//   'name' is the name of the binding
//   'isbinary' should be true if the data in local_filename is not UTF-8
//       (or compatible, like ASCII).
//   'filename' is the original name of the file this data was obtained from
//   'value' is a null-terminated string to bound to name.
//      Note that while it is valid to set 'isbinary' for a value that is
//        a null-terminated string, it probably is not correct.
//
int
reportfile_add_binding_from_memory( reportfile_t* file, 
                                    const char* name, 
                                    const void* value,
                                    size_t size,
                                    int isbinary, 
                                    const char* filename ) {
  if (size == SIZE_MAX)
    {
      internal_error_printf(
          "reportfile_add_binding_from_memory: error 'size' too long");
      return -1;
    }
  return reportfile_add_binding_from_region(file, name, 
                                            value, size, 
                                            isbinary, filename);
}



//
// Add a new binding to a report, for a binding whos value is represented
//   as a named file.
//
//   'file' is the reportfile to add the binding to.
//   'name' is the name of the binding
//   'isbinary' should be true if the data in local_filename is not UTF-8
//       (or compatible, like ASCII).
//   'filename' is the original name of the file this data was obtained from
//   
//   'local_filename' is the name of the file containing the bindings value
//      'filename' and 'local_filename' can be the same, but don't need to
//         be the same.
//
int
reportfile_add_binding_from_namedfile( reportfile_t* file, const char* name, const char* local_filename, int isbinary, const char* filename ) {
  int rc;

  rc = internal_reportfile_start_binding(file, name, isbinary, filename);
  if (rc < 0)
    return rc;

  char* content_dir_name = concat_path_file(file->tmpdir, "content");
  if (!content_dir_name)
    return -1;
  char* ofile_name = concat_path_file(content_dir_name, name);
  if (!ofile_name)
    return -1;
  rc = copy_file(local_filename, ofile_name);
  if (rc < 0)
    return rc;
  free(content_dir_name);
  free(ofile_name);

  
  char* href_name = concat_path_file("content", name);
  if (!href_name)
    return -1;
  rc = xmlTextWriterWriteAttribute(file->writer, (const xmlChar*)"href", (const xmlChar*)href_name);
  if (rc < 0) {
    internal_error_printf
	("strata_client: Error at xmlTextWriterWriteAttribute\n");
      return -1;
    }

  free(href_name);
  return 0;
}

//
// End the reportfile, and prepare it for delivery. 
// No more bindings can be added after this.
//
int
reportfile_end( reportfile_t* file ) {

  int rc;

  // close off the end of the xml file
  rc = xmlTextWriterEndDocument(file->writer);
  if (rc < 0) {
    internal_error_printf("strata_client: Error at xmlTextWriterEndDocument\n");
    return -1;
  }

  xmlFreeTextWriter(file->writer);
  file->writer = NULL;

  return 0;
}

//
// Write the reportfile to 'outfile_name'
//
// Currently this only works if the reportfile is a Report.
//
int
reportfile_write_to_file( reportfile_t* file, const char* outfile_name ) {

  if (0) {
    internal_error_printf("error: reportfile_write_to_file called before reportfile_end");
    return -1;
  }

  if (file->issignature) {
    internal_error_printf("error: reportfile_write_to_file called on signature");
    return -1;
  }



  int rc;
  // write the contents of the xmlfile to 'content.xml'
  rc = write_rgn_to_file((const char*)file->buf->content, SIZE_MAX, 
                         file->tmpdir, "content.xml");
  if (rc < 0)
    return rc;

  // tar up the contents
  char* cmd = ssprintf("tar --file=%s -C %s --create --gzip content content.xml", outfile_name, file->tmpdir);
  if (!cmd) 
    return -1;
  rc = run_command(cmd);
  if (rc < 0)
    return -1;
  free(cmd);

  return 0;
}


//
// Return the contents of the reportfile as a string.
//
// Currently only works if 'file' is a signature.
//
char* 
reportfile_as_string( reportfile_t* file ) {

  if (0) {
    internal_error_printf("error: reportfile_as_string called before reportfile_end");
    return NULL;
  }

  if (!file->issignature) {
    internal_error_printf("error: reportfile_as_string called on non-signature");
    return NULL;
  }

  return strdup((const char*)file->buf->content);
}



//
// free the reportfile
//
int
reportfile_free( reportfile_t* file ) {
  
  if (file) {
    int rc;
    char* cmd;

    // delete the temp dir
    cmd = ssprintf("rm -rf %s", file->tmpdir);
    if (!cmd) 
      return -1;
    rc = run_command(cmd);
    if (rc < 0)
      return -1;
    free(cmd);

    xmlFreeTextWriter(file->writer);
    xmlBufferFree(file->buf);
    free(file->tmpdir);
    free(file);
  }

  return 0;
}




static void
internal_error_curl(CURLcode err, const char* msg1, const char* msg2)
{
  internal_error_printf("error: %s: %s: %s", msg1, msg2, curl_easy_strerror(err));
}


typedef struct response_data {
  long code;
  const char* location;
  const char* strata_message;
  char* body;
  size_t body_size;
} response_data_t;

// 
// Examine each header looking for "Location:" or "Strata-Message:" headers
// 
static size_t 
headerfunction(void *buffer_pv, size_t count, size_t nmemb, void *response_data_pv)
{
  size_t size = count * nmemb;
  struct response_data* response_data = (struct response_data*)response_data_pv;
  const char* buffer = (const char*)buffer_pv;

  const char location_key[] = "Location:";
  const size_t location_key_size = sizeof(location_key)-1;

  if (size >= location_key_size
      && 0 == memcmp(buffer,location_key,location_key_size)) {
    const char* start = (const char*) buffer+location_key_size+1;
    const char* end;

    // skip over any leading space
    while (start < buffer+size 
           && isspace(*start)) ++start;

    end = start;

    // skip till we find the end of the url (first space or end of buffer)
    while (end < buffer+size 
           && !isspace(*end)) ++end;

    response_data->location = strndup(start,end-start);
  }


  const char strata_message_key[] = "Strata-Message:";
  const size_t strata_message_key_size = sizeof(strata_message_key)-1;

  if (size >= strata_message_key_size
      && 0 == memcmp(buffer,strata_message_key,strata_message_key_size)) {
    const char* tmp_msg = (const char*) buffer+strata_message_key_size;
    const char* tmp_end = (const char*) buffer+size-2;  // trim trailing \r\n

    while (tmp_msg < tmp_end && isspace(*tmp_msg)) 
      tmp_msg++;

    while (tmp_msg < tmp_end && isspace(*(tmp_end-1)))
      tmp_end--;

    size_t tmp_msg_len = tmp_end - tmp_msg;

    if (response_data->strata_message) {
      const char* old_msg = response_data->strata_message;
      const size_t old_msg_len = strlen(old_msg);
      const size_t new_msg_len = old_msg_len + tmp_msg_len + 1;
      char* new_msg = (char*)malloc(new_msg_len+1);

      mempcpy(mempcpy(mempcpy(new_msg, old_msg, old_msg_len), 
                              " ", 1), 
                              tmp_msg, tmp_msg_len);
      new_msg[new_msg_len] = 0;
      response_data->strata_message = new_msg;
      free((void*)old_msg);
    } else {
      response_data->strata_message = strndup(tmp_msg, tmp_msg_len);
    }
  }

  return size;
}

static int
append_accept_language_header(struct curl_slist ** httpheader_list_ptr) {

  // get the current MESSAGES locale
  char* environ_locale = setlocale(LC_MESSAGES, NULL);


  // catch all the locales we don't care about
  if (environ_locale == NULL 
      || 0 == strcmp("", environ_locale)
      || 0 == strcmp("C", environ_locale)
      || 0 == strcmp("POSIX", environ_locale))
    return 0;


  // Translate the environ locale to an http locale
  //   strip off everything after the initial list of alpha-numeric-underscores
  //   change underscores to dashes
  char* message_locale = strdup(environ_locale);
  if (!message_locale) {
    internal_error_printf("strdup failed");
    return -1;
  }
  char* p = NULL;
  for (p = message_locale; *p != 0 && (*p == '_' || isalnum(*p)); p++) {
    if (*p == '_')
      *p = '-';
  }
  *p = 0;


  // append the header to the header list
  char* header = ssprintf("Accept-Language: %s", message_locale);
  if (!header)
    return -1;
  *httpheader_list_ptr = curl_slist_append(*httpheader_list_ptr, header);


  free(header);
  free(message_locale);
  return 0;
}
  

  
    
//
// URL: to post to
//
// content_type: of the data
// data: depends on data_size
//   if data_size is 0 or more
//      data is binary data to be sent in post
//   if data_size == -1 
//      data is NULL terminated string to be sent in post
//   if data_size == -2 
//      data is name of file whose contents are to be sent in post
//   if data_size == -3 
//      data is name of file whose contents are to be sent in post
//      the file is sent as a multipart/form-data post
//


static response_data_t* 
post(const char* URL, 
     const char* username,
     const char* password,
     const char* content_type,
     const char* data,
     const long data_size) {

  CURLcode rc;
  CURL *handle = curl_easy_init();
  //  FILE* file;
  response_data_t* response_data = (response_data_t*)calloc(1,sizeof(response_data_t));

  rc = curl_easy_setopt(handle, CURLOPT_VERBOSE, 0);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_VERBOSE)");
    return NULL;
  }
  rc = curl_easy_setopt(handle, CURLOPT_NOPROGRESS, 1);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_NOPROGRESS)");
    return NULL;
  }


  rc = curl_easy_setopt(handle, CURLOPT_POST, 1);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_POST)");
    return NULL;
  }

  rc = curl_easy_setopt(handle, CURLOPT_URL, URL);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_URL)");
    return NULL;
  }

  if (username) {
    rc = curl_easy_setopt(handle, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
    if (rc) {
      internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_HTTPAUTH, CURLAUTH_BASIC)");
      return NULL;
    }
    
    rc = curl_easy_setopt(handle, CURLOPT_USERNAME, username);
    if (rc) {
      internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_USERNAME)");
      return NULL;
    }
    
    rc = curl_easy_setopt(handle, CURLOPT_PASSWORD, (password ? password : ""));
    if (rc) {
      internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_PASSWORD)");
      return NULL;
    }
  }
  
  FILE* data_file = 0;
  struct curl_httppost* post = NULL;  
  struct curl_httppost* last = NULL;  

  if (data_size == -2) {
    data_file = fopen(data, "r");
    if (!data_file) {
      internal_error_printf("Can't open file '%s'", data);
      return NULL;
    }
    rc = curl_easy_setopt(handle, CURLOPT_READDATA, data_file);
    if (rc) {
      internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_READDATA)");
      return NULL;
    }

  } else if (data_size == -3) {
    rc = curl_formadd(&post, &last, 
                      CURLFORM_PTRNAME, "file", 
                      CURLFORM_FILE, data, 
                      CURLFORM_CONTENTTYPE, content_type,
                      CURLFORM_FILENAME, data,
                      CURLFORM_END);
    if (rc) {
      internal_error_curl(rc, "problem", "curl_formadd(CURLFORM_FILE)");
      return NULL;
    }
    
    rc = curl_easy_setopt(handle, CURLOPT_HTTPPOST, post);
    if (rc) {
      internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_HTTPPOST);");
      return NULL;
    }

  } else {
    rc = curl_easy_setopt(handle, CURLOPT_POSTFIELDS, data);
    if (rc) {
      internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_POSTFIELDS)");
      return NULL;
    }
    rc = curl_easy_setopt(handle, CURLOPT_POSTFIELDSIZE, data_size);
    if (rc) {
      internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_POSTFIELDSIZE)");
      return NULL;
    }
  }


  struct curl_slist *httpheader_list = NULL;

  if (data_size != -3) {
    char* content_type_header = ssprintf("Content-Type: %s", content_type);
    if (!content_type_header) return NULL;
    httpheader_list = curl_slist_append(httpheader_list,content_type_header);
    free(content_type_header);
  }

  rc = append_accept_language_header(&httpheader_list);
  if (rc) return NULL;

  curl_easy_setopt(handle, CURLOPT_HTTPHEADER, httpheader_list);


  rc = curl_easy_setopt(handle, CURLOPT_HEADERFUNCTION, headerfunction);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_HEADERFUNCTION)");
    return NULL;
  }

  rc = curl_easy_setopt(handle, CURLOPT_WRITEHEADER, response_data );
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_WRITEHEADER)");
    return NULL;
  }

  FILE* body_stream = open_memstream(&response_data->body, &response_data->body_size); 
  rc = curl_easy_setopt(handle, CURLOPT_WRITEDATA, body_stream);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_WRITEDATA)");
    return NULL;
  }

  rc = curl_easy_perform(handle);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_perform");
    return NULL;
  }

  rc = curl_easy_getinfo(handle, CURLINFO_RESPONSE_CODE, &response_data->code); 
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_getinfo(CURLINFO_RESPONSE_CODE)");
    return NULL;
  }
 
  fclose(body_stream);
  if (data_file) fclose(data_file);
  if (post) curl_formfree(post);
  curl_easy_cleanup(handle);
  curl_slist_free_all(httpheader_list);

  return response_data;
}

static response_data_t* 
post_string(const char* URL,
            const char* username,
            const char* password,
            const char* content_type,
            const char* str) {
  return post(URL, username, password, content_type, str, -1);
}

static response_data_t* 
post_memory(const char* URL,
            const char* username,
            const char* password,
            const char* content_type,
            const char* data,
            const long data_size) {
  if (data_size < 0) {
    internal_error_printf("data size is negative");
    return NULL;
  }

  return post(URL, username, password, content_type, data, data_size);
}

static response_data_t* 
post_namedfile(const char* URL,
               const char* username,
               const char* password,
               const char* content_type,
               const char* namedfile) {
  return post(URL, username, password, content_type, namedfile, -2);
}

static response_data_t* 
postform_namedfile(const char* URL,
                   const char* username,
                   const char* password,
                   const char* content_type,
                   const char* namedfile) {
  return post(URL, username, password, content_type, namedfile, -3);
}



const char* 
post_signature(const char* baseURL, 
               const char* username,
               const char* password,
               const char* signature) {
  const char* URL = concat_path_file(baseURL, "/signatures");

  response_data_t* response_data = post_string(URL,
                                               username,
                                               password,
                                               "application/xml",
                                               signature);
  if (!response_data)
    return NULL;

  const char* retval;
  switch (response_data->code) {
  case 200:
  case 201:
    if (response_data->body && strlen(response_data->body) > 0) {
      retval = response_data->body;
      response_data->body = NULL;
    }
    else if (response_data->strata_message 
             && strcmp(response_data->strata_message,"CREATED") != 0) {
      retval = response_data->strata_message;
      response_data->strata_message = NULL;
    }
    else
      retval = strdup("Thank you for submitting your signature.");
    break;
  default:
    if (response_data->strata_message) 
      retval = ssprintf("Error: %s (http code %ld)", 
                        response_data->strata_message, 
                        response_data->code );
    else
      retval = ssprintf("Error: Response Code: %ld\nBody:\n%s", response_data->code, response_data->body);
  }

  free((void*)response_data->strata_message);
  free((void*)response_data->body);
  free((void*)response_data->location);
  free((void *)response_data);
  free((void *)URL);
  return retval;
}

static const char* 
make_case_data(const char* summary, const char* description,
               const char* product, const char* version,
               const char* component) {
  const char* retval;
  xmlTextWriterPtr writer;
  xmlBufferPtr buf;
  int rc;

  buf = xmlBufferCreate();
  if (buf == NULL) {
    internal_error_printf("strata_client: Error creating the xml buffer\n");
    return NULL;
  }
  writer = xmlNewTextWriterMemory(buf, 0);
  if (writer == NULL) {
    internal_error_printf("strata_client: Error creating the xml writer\n");
    return NULL;
  }

  // start a new xml document
  rc = xmlTextWriterStartDocument(writer, NULL, "UTF-8", "yes");
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterStartDocument\n");
    return NULL;
  }

  // start the document element 'case'
  rc = xmlTextWriterStartElement(writer, (const xmlChar*)"case");
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterStartElement\n");
    return NULL;
  }

  // add the 'xmlns' attribue
  rc = xmlTextWriterWriteAttribute(writer, (const xmlChar*)"xmlns",
                                   (const xmlChar*)"http://www.redhat.com/gss/strata");
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterWriteAttribute\n");
    return NULL;
  }

  // open a binding element
  rc = xmlTextWriterWriteElement(writer, (const xmlChar*)"summary", (const xmlChar*)summary);
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterWriteElement\n");
    return NULL;
  }

  rc = xmlTextWriterWriteElement(writer, (const xmlChar*)"description", (const xmlChar*)description);
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterWriteElement\n");
    return NULL;
  }

  if (product) {
    rc = xmlTextWriterWriteElement(writer, (const xmlChar*)"product", (const xmlChar*)product);
    if (rc < 0) {
      internal_error_printf
        ("strata_client: Error at xmlTextWriterWriteElement\n");
      return NULL;
    }
  }

  if (version) {
    rc = xmlTextWriterWriteElement(writer, (const xmlChar*)"version", (const xmlChar*)version);
    if (rc < 0) {
      internal_error_printf
        ("strata_client: Error at xmlTextWriterWriteElement\n");
      return NULL;
    }
  }

  if (component) {
    rc = xmlTextWriterWriteElement(writer, (const xmlChar*)"component", (const xmlChar*)component);
    if (rc < 0) {
      internal_error_printf
        ("strata_client: Error at xmlTextWriterWriteElement\n");
      return NULL;
    }
  }

  // close off the end of the xml file
  rc = xmlTextWriterEndDocument(writer);
  if (rc < 0) {
    internal_error_printf("strata_client: Error at xmlTextWriterEndDocument\n");
    return NULL;
  }

  retval = strdup((const char*)buf->content);
  xmlFreeTextWriter(writer);
  xmlBufferFree(buf);

  return retval;
}


static const char* 
make_response(const char* title, const char* body, 
              const char* actualURL, const char* displayURL) {
  const char* retval;
  xmlTextWriterPtr writer;
  xmlBufferPtr buf;
  int rc;
 
  buf = xmlBufferCreate();
  if (buf == NULL) {
    internal_error_printf("strata_client: Error creating the xml buffer\n");
    return NULL;
  }
  writer = xmlNewTextWriterMemory(buf, 0);
  if (writer == NULL) {
    internal_error_printf("strata_client: Error creating the xml writer\n");
    return NULL;
  }
  
  // start a new xml document
  rc = xmlTextWriterStartDocument(writer, NULL, "UTF-8", "yes");
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterStartDocument\n");
    return NULL;
  }

  rc = xmlTextWriterStartElement(writer, (const xmlChar*)"response");
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterStartElement\n");
    return NULL;
  }

  if (title) {
    rc = xmlTextWriterWriteElement(writer, (const xmlChar*)"title", (const xmlChar*)title);
    if (rc < 0) {
      internal_error_printf
        ("strata_client: Error at xmlTextWriterWriteElement\n");
      return NULL;
    }
  }

  if (body) {
    rc = xmlTextWriterWriteElement(writer, (const xmlChar*)"body", (const xmlChar*)body);
    if (rc < 0) {
      internal_error_printf
        ("strata_client: Error at xmlTextWriterWriteElement\n");
      return NULL;
    }
  }
    
  if (actualURL || displayURL) {
    rc = xmlTextWriterStartElement(writer, (const xmlChar*)"URL");
    if (rc < 0) {
      internal_error_printf
        ("strata_client: Error at xmlTextWriterStartElement\n");
      return NULL;
    }

    if (actualURL) {
      rc = xmlTextWriterWriteAttribute(writer, (const xmlChar*)"href", (const xmlChar*)actualURL);
      if (rc < 0) {
        internal_error_printf
          ("strata_client: Error at xmlTextWriterWriteAttribute\n");
        return NULL;
      }
    }

    if (displayURL) {
      rc = xmlTextWriterWriteString(writer, (const xmlChar*)displayURL);
      if (rc < 0) {
        internal_error_printf
          ("strata_client: Error at xmlTextWriterWriteAttribute\n");
        return NULL;
      }
    }
  }
  
  // close off the end of the xml file
  rc = xmlTextWriterEndDocument(writer);
  if (rc < 0) {
    internal_error_printf("strata_client: Error at xmlTextWriterEndDocument\n");
    return NULL;
  }

  retval = strdup((const char*)buf->content);
  xmlFreeTextWriter(writer);
  xmlBufferFree(buf);
  
  return retval;
}



const char* 
send_report_to_new_case( const char* baseURL,
                         const char* username,
                         const char* password,
                         const char* summary,
                         const char* description,
                         const char* component,
                         const char* report_file_name ) {

  const char* URL = concat_path_file(baseURL, "/cases");
  if (!URL) return NULL;

  const char* case_data = make_case_data(summary, description,
                                         "Red Hat Enterprise Linux", "6.0",
                                         component);
  if (!case_data) return NULL;

  response_data_t* createcase_response = post_string(URL,
                                                     username,
                                                     password,
                                                     "application/xml",
                                                     case_data);
  if (!createcase_response) return NULL;

  const char* retval = NULL;
  switch (createcase_response->code) {
  case 200:
  case 201:
    if (createcase_response->location 
        && strlen(createcase_response->location) > 0) {
      const char* attachURL = concat_path_file(createcase_response->location, 
                                               "/attachments");
      if (!attachURL) return NULL;

      response_data_t* attach_response = 
        postform_namedfile(attachURL,
                           username,
                           password,
                           "application/binary",
                           report_file_name);
      if (!attach_response) return NULL;

      switch (attach_response->code) {
      case 200:
      case 201:
        if (attach_response->body && strlen(attach_response->body) > 0) {
          if (createcase_response->body 
              && strlen(createcase_response->body) > 0) {
            const char* body_body = ssprintf("%s\n%s",
                                             createcase_response->body, 
                                             attach_response->body);
            retval = make_response("Case Created and Report Attached",
                                   body_body,
                                   createcase_response->location,
                                   "New Case URL");
            free((void*)body_body);
          } else {
            retval = make_response("Case Created and Report Attached",
                                   attach_response->body, 
                                   createcase_response->location,
                                   "New Case URL");
          }
        } else {
          if (createcase_response->body 
              && strlen(createcase_response->body) > 0) {
            retval = make_response("Case Created and Report Attached",
                                   createcase_response->body, 
                                   createcase_response->location,
                                   "New Case URL");
          } else {
            retval = make_response("Case Created and Report Attached",
                                   NULL, 
                                   createcase_response->location,
                                   "New Case URL");
          }
        }
        break;
      default:
        /* Case Creation Succeeded, attachement FAILED */
        if (createcase_response->body 
            && strlen(createcase_response->body) > 0) {
          if (attach_response->strata_message 
              && strlen(attach_response->strata_message) > 0)  {
            const char* body_body = ssprintf("%s\n%s",
                                             createcase_response->body, 
                                             attach_response->strata_message);
            retval = make_response("Case Created but Report Attachment Failed",
                                   body_body,
                                   createcase_response->location,
                                   "New Case URL");
            free((void*)body_body);
          } else if (attach_response->body 
                     && strlen(attach_response->body) > 0) {
            const char* body_body = ssprintf("%s\n%s",
                                             createcase_response->body, 
                                             attach_response->body);
            retval = make_response("Case Created but Report Attachment Failed",
                                   body_body,
                                   createcase_response->location,
                                   "New Case URL");
            free((void*)body_body);
          } else {
            const char* body_body = ssprintf("%s\n"
               "Report Attachement Failed with HTTP Code %ld",
                                             createcase_response->body, 
                                             attach_response->code);
            retval = make_response("Case Created but Report Attachment Failed",
                                   body_body,
                                   createcase_response->location,
                                   "New Case URL");
            free((void*)body_body);
          } 
        } else {
          if (attach_response->strata_message 
              && strlen(attach_response->strata_message) > 0)  {
            retval = make_response("Case Created but Report Attachment Failed",
                                   attach_response->strata_message, 
                                   createcase_response->location,
                                   "New Case URL");
          } else if (attach_response->body 
                     && strlen(attach_response->body) > 0) {
            retval = make_response("Case Created but Report Attachment Failed",
                                   attach_response->body,
                                   createcase_response->location,
                                   "New Case URL");
          } else {
            const char* body_body = ssprintf(
               "Report Attachement Failed with HTTP Code %ld",
               attach_response->code);
            retval = make_response("Case Created but Report Attachment Failed",
                                   body_body,
                                   createcase_response->location,
                                   "New Case URL");
            free((void*)body_body);
          }
        }
        break;
      }
      free((void*)attach_response->strata_message);
      free((void*)attach_response->body);
      free((void*)attach_response->location);
      free((void*)attach_response);
      free((void*)attachURL);
    } else {
      /* Case Creation returned valid code, but no location */
      retval = ssprintf("Error: case creation return HTTP Code %ld, "
                           "but no Location URL header", 
                        createcase_response->code);
    }
    break;
  default:
    if (createcase_response->strata_message 
        && strlen(createcase_response->strata_message) > 0)  {
      retval = make_response("Case Creation Failed", 
                             createcase_response->strata_message,
                             NULL, NULL);
    } else if (createcase_response->body 
               && strlen(createcase_response->body) > 0) {
      retval = make_response("Case Creation Failed", 
                             createcase_response->body,
                             NULL, NULL );
    } else {
      retval = ssprintf("Case Creation Failed with HTTP Code %ld", 
                        createcase_response->code );
    }
    break;
  }

  free((void*)case_data);
  free((void*)createcase_response->strata_message);
  free((void*)createcase_response->body);
  free((void*)createcase_response->location);
  free((void*)createcase_response);
  free((void*)URL);
  return retval;
}

const char* 
send_report_to_existing_case( const char* baseURL,
                              const char* username,
                              const char* password,
                              const char* case_name,
                              const char* report_file_name ) {

  const char* caseURL;
  const char* attachURL;
  const char* freethis;

  caseURL = concat_path_file(baseURL, "/cases");
  if (!caseURL) return NULL;

  freethis = caseURL;
  caseURL = concat_path_file(caseURL, case_name);
  if (!caseURL) return NULL;
  free((void*)freethis);

  attachURL = concat_path_file(caseURL, "/attachments");
  if (!attachURL) return NULL;
  
  response_data_t* attach_response = 
    postform_namedfile(attachURL, 
                       username, password,
                       "application/binary", report_file_name);
  if (!attach_response) return NULL;
  
  const char* retval = NULL;
  switch (attach_response->code) {
  case 200:
  case 201:
    if (attach_response->body && strlen(attach_response->body) > 0) {
      retval = make_response("Report Attached", 
                             attach_response->body,
                             caseURL,
                             "Case URL");
    } else {
      retval = make_response("Report Attached",
                             "Report was attached to case.",
                             caseURL,
                             "Case URL");
    }
    break;
  default:
    if (attach_response->strata_message 
        && strlen(attach_response->strata_message) > 0)  {
      retval = make_response("Report Attachement Failed",
                             attach_response->strata_message,
                             caseURL,
                             "Case URL");
    } else if (attach_response->body && strlen(attach_response->body) > 0) {
      retval = make_response("Report Attachement Failed",
                             attach_response->body,
                             caseURL,
                             "Case URL");
    } else {
      const char* body_body = ssprintf(
               "Report Attachement Failed with HTTP Code %ld",
               attach_response->code);
      retval = make_response("Report Attachement Failed",
                             body_body,
                             caseURL,
                             "Case URL");
      free((void*)body_body);
    }
  }
  
  free((void*)attachURL);
  free((void*)caseURL);
  free((void*)attach_response->strata_message);
  free((void*)attach_response->body);
  free((void*)attach_response->location);
  free((void *)attach_response);
  return retval;
}

